// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package cloud

import (
	"bufio"
	"errors"
	"fmt"
	"log"
	"os"
	"regexp"
	"sort"
	"strings"
	"time"

	"launchpad.net/snappy-tests-job/utils"
)

const (
	imgTpl                   = ".*%s-snappy-core-amd64-%s.*"
	listImagesCmd            = "nova image-list"
	launchInstanceCmdTpl     = "nova boot --key-name %s_%s --security-groups ssh --flavor %s --user-data /usr/share/snappy-tests-job/user.yml --image %s test-%d"
	describeInstanceCmdTpl   = "nova show %s"
	killInstanceCmdTpl       = "nova delete %s"
	maxAttempsInstanceWait   = 10
	secondsToWaitForInstance = 20
)

// CanonistackClient is the client for connecting to Canonistack
type CanonistackClient struct {
	util       utils.Utilizer
	flavor     string
	instanceID string
}

// CreateInstance launches a snappy instance from the latest image with the passed release
// and channel and returns its ip
func (c *CanonistackClient) CreateInstance(release, channel string) (ip string, err error) {
	log.Printf("*** Creating Snappy instance for release %s and channel %s ***", release, channel)
	imageID, err := c.listImages(release, channel)
	if err == nil {
		ip, err = c.launchInstance(imageID)
	}
	return
}

// KillInstance kills the current running instance if any
func (c *CanonistackClient) KillInstance() (err error) {
	if c.instanceID != "" {
		log.Printf("*** Killing instance %s ***", c.instanceID)
		err = c.killInstance()
	}
	return
}

// NewCanonistackClient is CanonistackClient's constructor
func NewCanonistackClient(flavor string, util utils.Utilizer) *CanonistackClient {
	return &CanonistackClient{flavor: flavor, util: util}
}

func (c *CanonistackClient) imgTemplate(release, channel string) string {
	fixedRelease := strings.Replace(release, ".", "", -1)
	return fmt.Sprintf(imgTpl, fixedRelease, channel)
}

func (c *CanonistackClient) listImages(release, channel string) (imageID string, err error) {
	params := utils.NewExecCommandParams("", strings.Fields(listImagesCmd), false)
	list, err := c.util.ExecCommand(params)
	if err == nil {
		imageID = c.getLatestImageID(release, channel, list)
		log.Printf("Found image with id %s for release %s and channel %s", imageID, release, channel)
	}
	return
}

func (c *CanonistackClient) launchInstance(imageID string) (ip string, err error) {
	log.Printf("Launching instance for Snappy image %s", imageID)

	launchInstanceCmd := fmt.Sprintf(launchInstanceCmdTpl,
		os.Getenv("NOVA_USERNAME"), os.Getenv("OS_REGION_NAME"), c.flavor, imageID, os.Getpid())
	params := utils.NewExecCommandParams("", strings.Fields(launchInstanceCmd), false)

	output, err := c.util.ExecCommand(params)
	if err == nil {
		c.instanceID = c.getinstanceID(output)
		log.Print("Instance launch requested, getting IP...")
		ip, err = c.getInstanceIP()
		if ip != "" {
			log.Printf("Got IP %s", ip)
		}
	}
	return
}

func (c *CanonistackClient) killInstance() (err error) {
	killInstanceCmd := fmt.Sprintf(killInstanceCmdTpl, c.instanceID)
	params := utils.NewExecCommandParams("", strings.Fields(killInstanceCmd), false)

	_, err = c.util.ExecCommand(params)
	return
}

func (c *CanonistackClient) getLatestImageID(release, channel, list string) (imageID string) {
	// list is of the form:
	//
	// | e37ef727-014d-48e8-b0e4-94cf39e1dbbc | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img                                   | ACTIVE |                                      |
	// | 633e3e02-bf9e-4de4-ba2b-361603a88233 | ubuntu-released/ubuntu-wily-alpha1-i386-server-20150624-disk1.img                                    | ACTIVE |                                      |
	// | 566e3e02-bf9e-4de4-ba2b-361603a88df6 | target-ami-id-20150110                            | ACTIVE |
	// | 566e3e02-bf9e-4de4-ba2b-361603a88df6 | target-ami-id-20150111                            | ACTIVE |
	// | 566e3e02-bf9e-4de4-ba2b-361603a88df6 | target-ami-id-20150109                            | ACTIVE |

	// In that case we should return target-ami-id-20150111
	reader := strings.NewReader(list)
	scanner := bufio.NewScanner(reader)
	scanner.Split(bufio.ScanLines)

	pattern := c.imgTemplate(release, channel)

	var imageIDs sort.StringSlice = []string{}
	for scanner.Scan() {
		line := scanner.Text()
		if match, _ := regexp.MatchString(pattern, line); match {
			fields := strings.Fields(line)
			imageIDs = append(imageIDs, fields[3])
		}
	}
	sort.Sort(sort.Reverse(imageIDs[:]))
	return imageIDs[0]
}

func (c *CanonistackClient) getinstanceID(list string) (instanceID string) {
	/*
		+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
			| Property                             | Value                                                                                                     |
			+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
			| OS-DCF:diskConfig                    | MANUAL                                                                                                    |
			| OS-EXT-AZ:availability_zone          | nova                                                                                                      |
			| OS-EXT-STS:power_state               | 0                                                                                                         |
			| OS-EXT-STS:task_state                | scheduling                                                                                                |
			| OS-EXT-STS:vm_state                  | building                                                                                                  |
			| OS-SRV-USG:launched_at               | -                                                                                                         |
			| OS-SRV-USG:terminated_at             | -                                                                                                         |
			| accessIPv4                           |                                                                                                           |
			| accessIPv6                           |                                                                                                           |
			| adminPass                            | HB5gEKXzp8Ls                                                                                              |
			| config_drive                         |                                                                                                           |
			| created                              | 2015-07-27T09:28:49Z                                                                                      |
			| flavor                               | m1.small (2)                                                                                              |
			| hostId                               |                                                                                                           |
			| id                                   | taget-instance-id
			| image                                | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img (e37ef727-014d-48e8-b0e4-94cf39e1dbbc) |
			| key_name                             | ues-snappy-integration-tests_lcy02                                                                        |
			| metadata                             | {}                                                                                                        |
			| name                                 | www                                                                                                       |
			| os-extended-volumes:volumes_attached | []                                                                                                        |
			| progress                             | 0                                                                                                         |
			| security_groups                      | ssh                                                                                                       |
			| status                               | BUILD                                                                                                     |
			| tenant_id                            | 245eb8a420d9466d919eb7ee3a54cc23                                                                          |
			| updated                              | 2015-07-27T09:28:49Z                                                                                      |
			| user_id                              | 0534127f70534300baf48e8ac1134734                                                                          |
			+--------------------------------------+-----------------------------------------------------------------------------------------------------------+

		 In that case we should return target-instance-id
	*/
	reader := strings.NewReader(list)
	scanner := bufio.NewScanner(reader)
	scanner.Split(bufio.ScanLines)

	for scanner.Scan() {
		line := scanner.Text()
		fields := strings.Fields(line)
		if len(fields) > 2 && fields[1] == "id" {
			instanceID = fields[3]
		}
	}
	return
}

func (c *CanonistackClient) getInstanceIP() (instanceIP string, err error) {
	describeInstanceCmd := fmt.Sprintf(describeInstanceCmdTpl, c.instanceID)

	params := utils.NewExecCommandParams("", strings.Fields(describeInstanceCmd), false)

	for i := 0; i < maxAttempsInstanceWait; i++ {
		output, _ := c.util.ExecCommand(params)

		if instanceIP = extractInstanceIP(output); instanceIP == "" {
			log.Print("IP not found, waiting...")

			time.Sleep(time.Duration(secondsToWaitForInstance) * time.Second)
		} else {

			log.Printf("The IP asigned to instance %s is %s", c.instanceID, instanceIP)
			break
		}
	}

	if instanceIP == "" {
		err = errors.New("Instance IP not found for instance " + c.instanceID)
	}
	return
}

func extractInstanceIP(output string) (instanceIP string) {
	/*
					   Output is of the form:
					   +--------------------------------------+-----------------------------------------------------------------------------------------------------------+
					   | Property                             | Value                                                                                                     |
					   +--------------------------------------+-----------------------------------------------------------------------------------------------------------+
					   | OS-DCF:diskConfig                    | MANUAL                                                                                                    |
					   | OS-EXT-AZ:availability_zone          | nova                                                                                                      |
					   | OS-EXT-STS:power_state               | 1                                                                                                         |
					   | OS-EXT-STS:task_state                | -                                                                                                         |
					   | OS-EXT-STS:vm_state                  | active                                                                                                    |
					   | OS-SRV-USG:launched_at               | 2015-07-27T09:29:27.000000                                                                                |
					   | OS-SRV-USG:terminated_at             | -                                                                                                         |
					   | accessIPv4                           |                                                                                                           |
					   | accessIPv6                           |                                                                                                           |
					   | <networkid> network                  | instance_ip                                                                                               |
					   | config_drive                         |                                                                                                           |
					   | created                              | 2015-07-27T09:28:49Z                                                                                      |
					   | flavor                               | m1.small (2)                                                                                              |
					   | hostId                               | f9b4d12d39c28f1f0ec970622c3c04348ab019b4b9571fe5567d76b6                                                  |
					   | id                                   | 4e4aff22-2e5e-42b5-b179-38ed9e6fb0e6                                                                      |
					   | image                                | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img (e37ef727-014d-48e8-b0e4-94cf39e1dbbc) |
					   | key_name                             | ues-snappy-integration-tests_lcy02                                                                        |
					   | metadata                             | {}                                                                                                        |
					   | name                                 | www                                                                                                       |
					   | os-extended-volumes:volumes_attached | []                                                                                                        |
					   | progress                             | 0                                                                                                         |
					   | security_groups                      | ssh                                                                                                       |
					   | status                               | ACTIVE                                                                                                    |
					   | tenant_id                            | 245eb8a420d9466d919eb7ee3a54cc23                                                                          |
					   | updated                              | 2015-07-27T09:29:27Z                                                                                      |
					   | user_id                              | 0534127f70534300baf48e8ac1134734                                                                          |
					   +--------------------------------------+-----------------------------------------------------------------------------------------------------------+
				  In this case, the ip should be instace_ip; <networkid> can be canonistack or, for neutron based stacks, a string containing the user name, as in
		      net_ues-snappy-integration

	*/

	reader := strings.NewReader(output)
	scanner := bufio.NewScanner(reader)
	scanner.Split(bufio.ScanLines)

	for scanner.Scan() {
		line := scanner.Text()
		fields := strings.Fields(line)

		if len(fields) > 3 && fields[2] == "network" {
			instanceIP = fields[4]
			break
		}
	}
	return
}
