// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package main

import (
	"flag"
	"log"
	"time"

	"launchpad.net/snappy-tests-job/cloud"
	"launchpad.net/snappy-tests-job/runner"
	"launchpad.net/snappy-tests-job/source"
	"launchpad.net/snappy-tests-job/utils"
)

const (
	defaultCloudProvider = "canonistack"
	defaultFlavor        = "m1.medium"
	defaultCommandTpl    = "go run ./_integration-tests/main.go -ip %s %s"
	defaultSrc           = "lp:snappy"
	defaultRelease       = "rolling"
	defaultChannel       = "edge"
)

func main() {
	var (
		useSnappyFromBranch = flag.Bool("snappy-from-branch", false,
			"If this flag is used, snappy will be compiled from this branch, copied to the testbed and used for the tests. Otherwise, the snappy installed with the image will be used.")
		arch = flag.String("arch", "",
			"Architecture of the test bed. Defaults to use the same architecture as the host.")
		outputDir = flag.String("output-dir", "", "Directory where test artifacts will be stored.")
		ip        = flag.String("ip", "",
			"IP of the testbed. If no IP is passed, a cloud instance will be started for the test.")
		cloudProvider = flag.String("cloud", defaultCloudProvider,
			"Cloud provider to use, defaults to "+defaultCloudProvider)
		flavor = flag.String("flavor", defaultFlavor,
			"Cloud flavor to use, defaults to "+defaultFlavor)
		release = flag.String("release", defaultRelease,
			"Release of the image, defaults to "+defaultRelease)
		channel = flag.String("channel", defaultChannel,
			"Channel of the image, defaults to "+defaultChannel)
		src = flag.String("src", defaultSrc,
			"Branch from which to get the source code, defaults to "+defaultSrc)
		target = flag.String("target", "",
			"Target branch to be merged with the src branch, defaults to not merging src with any branch")

		commandTpl = flag.String("command", defaultCommandTpl,
			"Command template, should include the testBed ip as %s, defaults to '"+defaultCommandTpl+"'")
	)

	flag.Parse()

	// TODO: we need a waitGroup here to run things more efficiently
	utilHandler := utils.NewBasicHandler()
	sourceHandler := source.NewBzrHandler(utilHandler)
	var err error
	var sourcePath string
	if *target == "" {
		sourcePath, err = sourceHandler.Get(*src)
	} else {
		sourcePath, err = sourceHandler.Merge(*src, *target)
	}
	if err != nil {
		log.Panicf("Error getting %s into %s", *src, sourcePath)
	}
	defer sourceHandler.CleanUp()

	ipString := *ip
	if ipString == "" {
		cloudClient := cloud.NewCloudClient(*cloudProvider, *flavor, utilHandler)
		ipString, err = cloudClient.CreateInstance(*release, *channel)
		defer cloudClient.KillInstance()

		if err != nil {
			log.Panicf("Error creating image for release %s and channel %s, %s", *release, *channel, err)
		}

		// give time for the instance to finish booting
		// TODO: check with a ticker if the instance is available
		log.Print("Waiting for instance to finish booting...")
		time.Sleep(3 * time.Minute)
	}

	testRunner := runner.NewBasicRunnerHandler(utilHandler)
	if err := testRunner.RunTests(sourcePath, *commandTpl, ipString, *arch, *outputDir, *useSnappyFromBranch); err != nil {
		log.Panic("Error from test runner")
	}
}
