// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package source

import (
	"log"
	"os"
	"path/filepath"
	"strconv"

	"launchpad.net/snappy-tests-job/utils"
)

const (
	targetPath   = "/tmp/snappy-tests-job"
	basePkg      = "src/launchpad.net"
	targetSubdir = "snappy"
)

// Sourcer is the interface satisfied by all the source handlers
type Sourcer interface {
	Get(string) (string, error)
	Merge(string, string) (string, error)
	CleanUp() error
}

// BzrHandler is a source handler for bazaar
type BzrHandler struct {
	util utils.Utilizer
}

// Get is the method for getting the source code
func (handler *BzrHandler) Get(repo string) (string, error) {
	path := filepath.Join(targetGoPath(), basePkg)

	log.Printf("*** Branching %s in %s ***", repo, path)

	err := handler.util.PrepareTargetDir(path)

	if err == nil {
		params := utils.NewExecCommandParams(path, []string{"bzr", "branch", repo, targetSubdir}, true)
		_, err = handler.util.ExecCommand(params)
	}

	return filepath.Join(path, targetSubdir), err
}

// Merge merges two branches and returns the path to the merged source.
func (handler *BzrHandler) Merge(source string, target string) (string, error) {
	path, err := handler.Get(target)
	if err == nil {
		log.Printf("*** Merging %s with %s ***", target, source)
		params := utils.NewExecCommandParams(path, []string{"bzr", "merge", source}, true)
		_, err = handler.util.ExecCommand(params)
	}
	return path, err
}

// NewBzrHandler returns a pointer to a new bzr source handler
func NewBzrHandler(util utils.Utilizer) *BzrHandler {
	return &BzrHandler{util: util}
}

func targetGoPath() string {
	return filepath.Join(targetPath, strconv.Itoa(os.Getpid()))
}

// CleanUp removes the directory used for setting up the
// target branch Go environment
func (handler *BzrHandler) CleanUp() (err error) {
	path := targetGoPath()
	return handler.util.RemoveAll(path)
}
